@@ -9,6 +9,9 @@ from einops import rearrange
 def exists(val):
     return val is not None
 
+def cast_tuple(val, repeat = 1):
+    return val if isinstance(val, tuple) else ((val,) * repeat)
+
 # sin activation
 
 class Sine(nn.Module):
@@ -53,13 +56,16 @@ class Siren(nn.Module):
 class SirenNet(nn.Module):
     def __init__(self, dim_in, dim_hidden, dim_out, num_layers, w0 = 1., w0_initial = 30., use_bias = True, final_activation = None):
         super().__init__()
-        layers = []
+        self.num_layers = num_layers
+        self.dim_hidden = dim_hidden
+
+        self.layers = nn.ModuleList([])
         for ind in range(num_layers):
             is_first = ind == 0
             layer_w0 = w0_initial if is_first else w0
             layer_dim_in = dim_in if is_first else dim_hidden
 
-            layers.append(Siren(
+            self.layers.append(Siren(
                 dim_in = layer_dim_in,
                 dim_out = dim_hidden,
                 w0 = layer_w0,
@@ -67,19 +73,50 @@ class SirenNet(nn.Module):
                 is_first = is_first
             ))
 
-        self.net = nn.Sequential(*layers)
-
         final_activation = nn.Identity() if not exists(final_activation) else final_activation
         self.last_layer = Siren(dim_in = dim_hidden, dim_out = dim_out, w0 = w0, use_bias = use_bias, activation = final_activation)
 
-    def forward(self, x):
-        x = self.net(x)
+    def forward(self, x, mods = None):
+        mods = cast_tuple(mods, self.num_layers)
+
+        for layer, mod in zip(self.layers, mods):
+            x = layer(x)
+
+            if exists(mod):
+                x *= rearrange(mod, 'd -> () d')
+
         return self.last_layer(x)
 
+# modulatory feed forward
+
+class Modulator(nn.Module):
+    def __init__(self, dim_in, dim_hidden, num_layers):
+        super().__init__()
+        dim = dim_in
+        self.layers = nn.ModuleList([])
+
+        for ind in range(num_layers):
+            is_first = ind == 0
+            dim = dim_in if is_first else dim_hidden
+
+            self.layers.append(nn.Sequential(
+                nn.Linear(dim, dim_hidden),
+                nn.ReLU()
+            ))
+
+    def forward(self, x):
+        hiddens = []
+
+        for layer in self.layers:
+            x = layer(x)
+            hiddens.append(x)
+
+        return tuple(hiddens)
+
 # wrapper
 
 class SirenWrapper(nn.Module):
-    def __init__(self, net, image_width, image_height):
+    def __init__(self, net, image_width, image_height, latent_dim = None):
         super().__init__()
         assert isinstance(net, SirenNet), 'SirenWrapper must receive a Siren network'
 
@@ -87,16 +124,30 @@ class SirenWrapper(nn.Module):
         self.image_width = image_width
         self.image_height = image_height
 
+        self.modulator = None
+        if exists(latent_dim):
+            self.modulator = Modulator(
+                dim_in = latent_dim,
+                dim_hidden = net.dim_hidden,
+                num_layers = net.num_layers
+            )
+
         tensors = [torch.linspace(-1, 1, steps = image_width), torch.linspace(-1, 1, steps = image_height)]
         mgrid = torch.stack(torch.meshgrid(*tensors), dim=-1)
         mgrid = rearrange(mgrid, 'h w c -> (h w) c')
         self.register_buffer('grid', mgrid)
 
-    def forward(self, img = None):
+    def forward(self, img = None, *, latent = None):
+        modulate = exists(self.modulator)
+        assert not (modulate ^ exists(latent)), 'latent vector must be only supplied if `latent_dim` was passed in on instantiation'
+
+        mods = self.modulator(latent) if modulate else None
+
         coords = self.grid.clone().detach().requires_grad_()
-        out = self.net(coords)
+        out = self.net(coords, mods)
         out = rearrange(out, '(h w) c -> () c h w', h = self.image_height, w = self.image_width)
 
         if exists(img):
             return F.mse_loss(img, out)
+
         return out
